/*
 * CastRegistry.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _CAST_REGISTRY_HPP_
#define _CAST_REGISTRY_HPP_

#include <memory>
#include <string>
#include <set>

#include <exception/CommonException.h>
#include <abstraction/OperationAbstraction.hpp>

#include <core/namingApi.hpp>

namespace abstraction {

class CastRegistry {
	class Entry {
		bool m_normalize;
	public:
		Entry ( bool normalize ) : m_normalize ( normalize ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const = 0;

		bool getNormalize ( ) const {
			return m_normalize;
		}

	};

	template < class Return, class Param >
	class DefaultEntryImpl : public Entry {
	public:
		DefaultEntryImpl ( bool normalize ) : Entry ( normalize ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	template < class Return, class Param >
	class AlgorithmEntryImpl : public Entry {
		std::function < Return ( Param ) > m_callback;
	public:
		AlgorithmEntryImpl ( std::function < Return ( Param ) > callback, bool normalize ) : Entry ( normalize ), m_callback ( callback ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	static ext::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > & getEntries ( ) {
		static ext::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > casts;
		return casts;
	};

public:
	template < class TargetType, class ParamType >
	static void registerCast ( std::string target, std::string param, bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new DefaultEntryImpl < TargetType, ParamType > ( normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCast ( bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCast < TargetType, ParamType > ( target, param, normalize );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( std::string target, std::string param, TargetType ( * callback ) ( const ParamType & ), bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new AlgorithmEntryImpl < TargetType, const ParamType & > ( callback, normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( TargetType ( * callback ) ( const ParamType & ), bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCastAlgorithm < TargetType, ParamType > ( target, param, callback, normalize );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( std::string target, std::string param, TargetType ( * callback ) ( ParamType ), bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new AlgorithmEntryImpl < TargetType, ParamType > ( callback, normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( TargetType ( * callback ) ( ParamType ), bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCastAlgorithm < TargetType, ParamType > ( target, param, callback, normalize );
	}

	static std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( const std::string & target, const std::string & param, bool & normalize ) {
		std::set < std::string > targetTypes;
		if ( alib::namingApi::hasTypes ( target ) )
			targetTypes = ext::transform < std::string > ( alib::namingApi::getTypes ( target ), [ ] ( const ext::type_index & type ) { return ext::to_string ( type ); } );
		else
			targetTypes.insert ( target );

		for ( const std::string & toType : targetTypes ) {
			auto cast = getEntries ( ).find ( std::make_pair ( toType, param ) );
			if ( cast != getEntries ( ).end ( ) ) {
				normalize = cast->second->getNormalize ( );
				return cast->second->getAbstraction ( );
			}
		}

		throw exception::CommonException ( "Entry from " + param + " to " + target + " not available." );
	}

	static bool isNoOp ( const std::string & target, const std::string & param ) {
		std::set < std::string > targetTypes;
		if ( alib::namingApi::hasTypes ( target ) )
			targetTypes = ext::transform < std::string > ( alib::namingApi::getTypes ( target ), [ ] ( const ext::type_index & type ) { return ext::to_string ( type ); } );
		else
			targetTypes.insert ( target );

		for ( const std::string & toType : targetTypes )
			if ( param == toType )
				return true;

		return false;
	}

	static bool castAvailable ( const std::string & target, const std::string & param ) {
		std::set < std::string > targetTypes;
		if ( alib::namingApi::hasTypes ( target ) )
			targetTypes = ext::transform < std::string > ( alib::namingApi::getTypes ( target ), [ ] ( const ext::type_index & type ) { return ext::to_string ( type ); } );
		else
			targetTypes.insert ( target );

		for ( const std::string & toType : targetTypes ) {
			auto cast = getEntries ( ).find ( std::make_pair ( toType, param ) );
			if ( cast != getEntries ( ).end ( ) )
				return true;
		}

		return false;
	}

	static std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( const std::string & target, const std::string & param ) {
		bool normalize;
		return getAbstraction ( target, param, normalize );
	}

	static void listFrom ( const std::string & type ) {
		std::set < std::string > sourceTypes;
		if ( alib::namingApi::hasTypes ( type ) )
			sourceTypes = ext::transform < std::string > ( alib::namingApi::getTypes ( type ), [ ] ( const ext::type_index & raw_type ) { return ext::to_string ( raw_type ); } );
		else
			sourceTypes.insert ( type );

		for ( const std::pair < const std::pair < std::string, std::string >, std::unique_ptr < Entry > > & entry : getEntries ( ) ) {
			if ( sourceTypes.count ( entry.first.first ) ) {
				std::cout << entry.first.second << std::endl;
			}
		}
	}

	static void listTo ( const std::string & type ) {
		std::set < std::string > targetTypes;
		if ( alib::namingApi::hasTypes ( type ) )
			targetTypes = ext::transform < std::string > ( alib::namingApi::getTypes ( type ), [ ] ( const ext::type_index & raw_type ) { return ext::to_string ( raw_type ); } );
		else
			targetTypes.insert ( type );

		for ( const std::pair < const std::pair < std::string, std::string >, std::unique_ptr < Entry > > & entry : getEntries ( ) ) {
			if ( targetTypes.count ( entry.first.second ) ) {
				std::cout << entry.first.first << std::endl;
			}
		}
	}

};

} /* namespace abstraction */

#include <abstraction/CastAbstraction.hpp>
#include <abstraction/AlgorithmAbstraction.hpp>

namespace abstraction {

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::DefaultEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::CastAbstraction < Return, const Param & > > ( );
}

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::AlgorithmEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::AlgorithmAbstraction < Return, const Param & > > ( m_callback );
}

} /* namespace abstraction */

#endif /* _CAST_REGISTRY_HPP_ */
